/*******************************************************************\

Module: A CNF Clause Container

Author: CM Wintersteiger

\*******************************************************************/

#include <errno.h>

#include <algorithm>
#include <iostream>
#include <sstream>

#include "cnf.h"

Literal::Literal( unsigned v, bool sign )
{
  set(v, sign);
}

signed long long Literal::to_dimacs( void ) const
{
  if (sign())
    return - ((signed long long) variable());
  else
    return variable();
}

Literal DClause::contains( unsigned v ) const 
{  
  for (unsigned i=0; i<size(); i++)
    if (content[i]==v)
      return content[i];
  return Literal(0,false);
}

bool DClause::contains( const Literal &l ) const
{
  for (unsigned i=0; i<size(); i++)
    if (content[i]==l)
      return true;
  return false;
}

void DClause::remove( unsigned v )
{
  unsigned i=0;
  while (i<size())
  {
    if (content[i]==v)
    {
      content[i] = content.back();
      content.pop_back();
      #ifdef __THERE_IS_ENOUGH_MEMORY
        count--;
      #endif
    }
    else 
      i++;    
  }
}

void DClause::remove( const Literal &l )
{
  unsigned i=0;
  while (i<size())
  {
    if (content[i]==l)
    {
      content[i] = content.back();
      content.pop_back();
      #ifdef __THERE_IS_ENOUGH_MEMORY
        count--;
      #endif
    }
    else 
      i++;    
  }
}

bool DClause::insert( const Literal &l )
{
  if (!contains(l))
  {
    content.resize(size()+1, l);
    #ifdef __THERE_IS_ENOUGH_MEMORY
      count++;
    #endif
    return true;
  }
  else
    return false;  
}

bool DClause::is_true( void ) const
{
  for (unsigned i=0; i<size(); i++)
  {
    for (unsigned j=i+1; i<size(); j++)    
      if (content[i].variable() == content[j].variable() &&
          content[i].sign() != content[j].sign())      
        return true;
  }
  return false;
}

bool DClause::empty( void ) const
{
  return content.empty();
}

void DClause::add( const DClause& c )
{
  content.reserve(size()+c.size());
  for (unsigned i=0; i<c.size(); i++)
    insert(c.content[i]);
}

void DClause::add_except( const DClause& c, unsigned exception )
{
  content.reserve(size()+c.size());
  for (unsigned i=0; i<c.size(); i++)
    if (c.content[i].variable()!=exception)
      insert(c.content[i]);
}

void DClause::clear( void )
{
  content.clear();
  #ifdef __THERE_IS_ENOUGH_MEMORY
    count = 0;
  #endif
}

void DClause::reduce( 
  const std::vector<unsigned> &levels,
  const std::vector<Quantifier> &quantifiers)
{  
  Quantifier::Type max_type = Quantifier::UNIVERSAL;
  unsigned max_ql = 1;
  unsigned max_var = 0;
  
  while (max_ql>0 && max_type == Quantifier::UNIVERSAL)
  {
    max_ql = 0;
    max_var = 0;
    
    for (unsigned i=0; i<size(); i++)
    {
      unsigned lvl = levels[content[i].variable()];
      if (lvl > max_ql) 
      {
        max_ql = lvl;
        max_type = quantifiers[content[i].variable()].type;
        max_var = content[i].variable();
      }      
    }    
    
    if (max_var!=0 && max_type==Quantifier::UNIVERSAL) 
    {
      //std::cout << "Reducing " << max_var << std::endl;
      remove(max_var);
    }
  }
}

bool DClause::qresolve_into( 
  const DClause &other, 
  unsigned pivot,
  const std::vector<unsigned> &levels,
  const std::vector<Quantifier> &quantifiers,
  DClause &into ) const
{
  Literal lc1 = contains(pivot);
  Literal lc2 = other.contains(pivot);
  
  into.clear();
  
  if (  lc1.variable()==lc2.variable() &&
        lc1.sign() != lc2.sign() )
  {
    into.add_except(*this, pivot);
    into.add_except(other, pivot);
    into.reduce(levels, quantifiers);
    //std::cout << " [" << size() << "]" << std::endl;
    return true;
  }
  return false;
}

std::string DClause::to_string( void ) const
{
  std::stringstream result;
  result << "(";
  unsigned i = 0;
  while (i<size())
  {
    result << content[i].to_dimacs();
    i++;
    if (i!=size())
      result << ", ";
  }
  result << ")";
  return result.str();
}

bool DClause::invert( CClause& cube ) const
{
  cube.clear();
  for (unsigned i=0; i<size(); i++)
  {
    Literal l = content[i];
    l.invert();
    cube.insert(l);
  }
  return true;
}

bool QCNF::qresolve( unsigned ci1, unsigned ci2, unsigned v)
{
  if ( ci1 >= size() || ci2 >= size() ) return false;  
     
  DClause c1 = at(ci1);
  DClause c2 = at(ci2);
  
//  std::cout << "Resolving [" << ci1+1 << "," << ci2+1 << "] " << c1.to_string() << 
//    " x " << c2.to_string() << " over " << v << std::endl;
  
  push_back(DClause());
  if ( c1.qresolve_into(  c2, v, 
                          quantification_levels, 
                          quantifiers, back()))
    return true;

  pop_back();  
  //std::cout << " [N/A]" << std::endl;  
  return false;
}

std::string CClause::to_string( void ) const
{
  std::stringstream result;
  result << "[";
  const_iterator it=content.begin();
  while (it!=content.end())
  {
    result << it->to_dimacs();
    it++;
    if (it!=content.end())
      result << ", ";
  }
  result << "]";
  return result.str();
}

bool CClause::is_false( void ) const
{
  for (const_iterator it=content.begin();
       it!=content.end();
       it++)
  {    
    for (const_iterator it2=it;
         it2!=content.end();
         it2++)
    {
      if (Literal(it->variable(), ! it->sign()) == *it2)
        return true;
    }
  }
  return false;
}

void CClause::add( const CClause& c )
{  
  for (const_iterator it=c.content.begin();
       it!=c.content.end();
       it++)
  {
    insert(*it);
  }
}

bool CClause::invert( DClause& clause ) const
{
  clause.clear();
  for (const_iterator it = content.begin();
       it!=content.end();
       it++)
  {
    Literal l = *it;
    l.invert();
    clause.insert(l);
  }
  return true;
}

void CNF::output( std::ostream& stream ) const
{
  stream << "[" << size() << "]";
  for (std::vector<DClause>::const_iterator it = begin();
       it!=end();
       it++)
  {
    stream << it->to_string();
  }
}

void DNF::output( std::ostream& stream ) const
{
  stream << "[" << size() << "]";
  for (std::vector<CClause>::const_iterator it = begin();
       it!=end();
       it++)
  {
    stream << it->to_string();
  }
}

bool CNF::invert( DNF& dnf ) const
{
  //std::cout << "DEBUG: Inverting CNF:";
  //output(std::cout); std::cout << std::endl;
  dnf.clear();
  
  for (const_iterator it = begin();
       it!=end();
       it++)
  {
    CClause t;
    if (!it->invert(t)) return false;
    dnf.push_back(t);
  }
  //std::cout << "DEBUG: Result DNF:";
  //dnf.output(std::cout); std::cout << std::endl;
  return true;
}

bool DNF::invert( CNF &cnf ) const
{
//  std::cout << "DEBUG: Inverting DNF:";
//  output(std::cout); std::cout << std::endl;
  cnf.clear();
  
  for (const_iterator it = begin();
       it!=end();
       it++)
  {
    DClause t;
    if (!it->invert(t)) return false;
    cnf.push_back(t);
  }
//  std::cout << "DEBUG: Result CNF:";
//  cnf.output(std::cout); std::cout << std::endl;
  return true;
}

/*******************************************************************\
 
Function: QCNF::read_qdimacs
 
  Inputs: stream
 
 Outputs: nothing
 
 Purpose: loads a qdimacs file into a cnf, throws exceptions on error.
 
\*******************************************************************/

bool QCNF::read_qdimacs( std::ifstream &in )
{
  if (!in) 
  {
    last_error = "QBF File not open.";
    return false;
  }
    
  std::string t;
  in >> t; 
  while (t=="c")
  {
    char c=0x0;
    while (c != 0xA) in.read(&c,1);
    in >> t;
  }
  
  if (t!="p") 
  {
    last_error = "QBF File does not start with a p line.";
    return false;
  }
  in >> t;
  if (t!="cnf") 
  {
    last_error = "QBF File does not have a cnf argument.";
    return false;
  }
  in >> t;
  errno = 0;
  max_variable_no = strtoul(t.c_str(), NULL, 10);
  if (errno!=0) 
  {
    last_error = "QBF File: Number of variables is bogus.";
    return false;
  }
  in >> t;
  errno = 0;
  unsigned noclauses = strtoul(t.c_str(), NULL, 10);
  if (errno!=0) 
  {
    last_error = "QBF File: Number of clauses is bogus.";
    return false;
  }
  reserve(noclauses);
  in >> t;
  Quantifier::Type cur_type = Quantifier::NONE;
  
  while (t=="e" || t=="a") 
  {
    if (t=="e") cur_type = Quantifier::EXISTENTIAL;
    else cur_type = Quantifier::UNIVERSAL;    
    errno = 0;
    unsigned varno=1;    
    while (varno!=0)
    {
      in >> t;
      varno = strtoul(t.c_str(), NULL, 10);
      if (errno!=0)
      {
        last_error = "QBF File: Invalid quantification sequence.";
        return false;
      }
      if (varno!=0)
      {        
        new_variable(varno, cur_type);
      }
    }
    in >> t;
  }
  
  DClause tcl;
  signed v;
  char *endp;
 
  errno = 0;
  while (in)
  {
    v = strtol(t.c_str(), &endp, 10);
    if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
    {
      last_error = "QBF File: Bogus clause.";
      return false;
    }
        
    if (v==0)
    {
      push_back(tcl);
      tcl.clear();
    }
    else if (v>0)
    {      
      tcl.insert(Literal(v, false));
    }
    else // v<0
    {
      tcl.insert(Literal(-v, true));
    }
    
    // quantify unquantified variables
    if (v!=0)
    {
      if (v<0) v=-v;
      if ((unsigned)v>max_variable_no) max_variable_no = v;
      if ((unsigned)v>=quantifiers.size()) quantifiers.resize(v+1); 
      if (quantifiers[v].type==Quantifier::NONE)
      {
        new_variable(v, Quantifier::EXISTENTIAL, true);      
      }
    }
      
    in >> t;
  }
  if (in) 
  {
    last_error = "QBF File: Error reading CNF.";
    return false;
  }
  
  return true;
}

void QCNF::write_qdimacs( std::ofstream& out ) const
{
  out << "p cnf " << max_variable_no << " " << size() << std::endl;
  for (std::vector<qscopet>::const_iterator it=quantification_sequence.begin();
       it!=quantification_sequence.end();
       it++)
  {
    for (qscopet::const_iterator sit = it->begin();
         sit!=it->end();
         sit++)
    {
      if (sit->type==Quantifier::EXISTENTIAL)
        out << "e " << sit->variable << " 0" << std::endl;
      else if (sit->type==Quantifier::UNIVERSAL)
        out << "a " << sit->variable << " 0" << std::endl;
    }
  }
  for (CNF::const_iterator it = begin();
       it!=end();
       it++)
  {
    for (DClause::const_iterator lit = it->begin();
         lit!=it->end();
         lit++)
    {
      out << lit->to_dimacs() << " ";
    }
    out << "0" << std::endl;
  }
}

bool QCNF::new_variable( unsigned varno, Quantifier::Type t, bool outer )
{  
  if (outer && t==Quantifier::UNIVERSAL)
  {
    last_error = "New outermost variables may not be universal.";
    return false;
  } 
  
  if (is_defined(varno))
  {
    last_error = "Variable redefined";
    return false;
  }
  
  if (varno>=quantifiers.size()) 
    quantifiers.resize(varno+1); 
  quantifiers[varno].type = t;
  quantifiers[varno].variable = varno;
  
  if (outer || quantification_sequence.size()==0)
  {
    if (varno>=quantification_levels.size())
      quantification_levels.resize(varno+1, 0);
    if (t==Quantifier::UNIVERSAL) 
    {
      quantification_levels[varno] = 1; // leave some space for outermost ex.
      if (quantification_sequence.size()<2)
        quantification_sequence.resize(2);
      quantification_sequence[1].push_back(quantifiers[varno]);
    }
    else
    {
      quantification_levels[varno] = 0;
      if (quantification_sequence.size()==0)
        quantification_sequence.resize(1);
      quantification_sequence[0].push_back(quantifiers[varno]);
    }
        
  }
  else
  {  
    qscopet& last = quantification_sequence.back();
    
    if (varno>=quantification_levels.size())
      quantification_levels.resize(varno+1, 0);
    if (last.back().type==t)
    {
      quantification_levels[varno] = 
        quantification_levels[last.back().variable];
      last.push_back(quantifiers[varno]);
    }
    else
    {
      quantification_levels[varno] = 
        quantification_levels[last.back().variable]+1;
      quantification_sequence.resize(quantification_sequence.size()+1);
      quantification_sequence.back().push_back(quantifiers[varno]);
    }
        
  }
  
  if (varno>max_variable_no) max_variable_no = varno;
  
//  std::cout << "DEBUG: new innermost variable #" << varno << ", ql=" <<
//    quantification_levels[varno] << std::endl;
  
  return true;
}

bool QCNF::new_variable_after( 
  unsigned varno, 
  Quantifier::Type t, 
  unsigned after)
{
  if (!is_defined(after))
  {
    std::stringstream s;
    s << "Definition depends on undefined variable #" << after; 
    last_error = s.str();
    return false;
  }
  if (is_defined(varno))
  {
    last_error = "Variable redefinition";
    return false;
  }
  
  unsigned newlvl = quantification_levels[after];  
  qscopet &scope = quantification_sequence[newlvl];
    
  Quantifier q;
  q.type = t;
  q.variable = varno;  
      
  if (scope.front().type==t)
  {    
    qscopet::iterator it = scope.begin();
    for (; it!=scope.end(); it++)
      if (it->variable==after) break;
    it++;
    scope.insert(it, q);
  } 
  else
  { 
    newlvl++;
    if (quantification_sequence.size()<=newlvl)
      quantification_sequence.resize(quantification_sequence.size()+1);
    quantification_sequence[newlvl].push_back(q);
  }
  
  if (varno>=quantification_levels.size())
    quantification_levels.resize(varno+1, 0);
  quantification_levels[varno] = newlvl;
  
  if (varno>=quantifiers.size()) 
    quantifiers.resize(varno+1); 
  quantifiers[varno] = q;
 
  if (varno>max_variable_no) max_variable_no = varno;
  
//  std::cout << "DEBUG: new variable #" << varno << ", ql=" <<
//    quantification_levels[varno] << std::endl;
  
  return true;
}

bool QCNF::is_defined( unsigned v ) const
{
   if (v < quantifiers.size() &&
       quantifiers[v].type!=Quantifier::NONE)
       return true;
   return false;       
}
