/*******************************************************************\

Module: QBV, The Quantified Boolean Validator

Author: CM Wintersteiger

\*******************************************************************/

#include <errno.h>

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>

#ifdef __USE_MINISAT2
  #include <SimpSolver.h>
#else
  #include <Solver.h>
#endif

#include "qbv.h"

/*******************************************************************\
 
Function: QBV::QBV
 
  Inputs: none
 
 Outputs: none
 
 Purpose: Constructor.
 
\*******************************************************************/

QBV::QBV( void ) 
{
  cur_linenumber=1;
  cur_colnumber=1;
  test_trace = false;
  debug = false;
}

/*******************************************************************\
 
Function: QBV::validate
 
  Inputs: char *, char *
 
 Outputs: true if the certificate was correct, false otherwise
 
 Purpose: validates a certificate file.
 
\*******************************************************************/

bool QBV::validate( const char *qbf, const char* qbc )
{
  std::string qbffn = qbf;
  std::string qbcfn = qbc;
  return validate(qbffn, qbcfn);  
}

/*******************************************************************\
 
Function: QBV::resolution_line
 
  Inputs: certificate file object
 
 Outputs: nothing
 
 Purpose: processes a resolution line
 
\*******************************************************************/

void QBV::resolution_line( std::ifstream &certificate_file )
{
  unsigned c1 = read_unsigned(certificate_file);
  unsigned c2 = read_unsigned(certificate_file);
  unsigned v = read_unsigned(certificate_file);
  if (!qcnf.qresolve(c1-1, c2-1, v))
    throw Exception("Resolution failed.");
}

/*******************************************************************\
 
Function: QBV::trace_line
 
  Inputs: certificate file object
 
 Outputs: nothing
 
 Purpose: processes a resolution line
 
\*******************************************************************/

void QBV::trace_line( unsigned newinx, std::ifstream &certificate_file )
{    
  Resolution resolution;
  
  resolution.newinx = newinx;
  
  if (newinx<clause_map.size() && clause_map[newinx]!=0)
  {
    std::stringstream s;
    s << "Redefinition of clause #" << newinx;
    throw Exception(s.str());
  }
  
  // parse resolvent
  while (certificate_file.peek()==' ') read_character(certificate_file);  
  if (certificate_file.peek()!='*') // we have a resolvent in there.
  {
    signed l = read_signed(certificate_file);
    while (l!=0)
    { 
      resolution.resolvent.insert(Literal( (l<0)?-l:l, (l<0)?true:false ));
      if (!qcnf.is_defined((l<0)?-l:l))
        qcnf.new_variable((l<0)?-l:l, Quantifier::EXISTENTIAL, true);
      l = read_signed(certificate_file);
    }
  }
  else
  {
    std::string& t = read_token(certificate_file);
    if (t=="*")
      resolution.star_resolvent = true;
    else
      throw Exception("Bogus resolvent.");
  }
  
  // parse antecedents
  unsigned a = read_unsigned(certificate_file);
  while (a!=0)
  {
    resolution.antecedents.push_back(a);
//    if (a>=clause_map.size() || clause_map[a]==0)
//    {
//      std::stringstream s;
//      s << "Antecedent references undefined clause (" << a << ").";
//      throw Exception(s.str());
//    }    
    a = read_unsigned(certificate_file);
  }
    
  work_list.push_front(resolution);
  
  bool did_something = true;
  
  do {
    did_something=false;    
    for (std::list<Resolution>::iterator it = work_list.begin();
         it!=work_list.end();)
    {
      Resolution &r = *it;
      bool all_antecedents_here = true;
      for (std::list<unsigned>::const_iterator ait = r.antecedents.begin();
           ait!=r.antecedents.end();
           ait++)
      {
        if (*ait>=clause_map.size() || clause_map[*ait]==0) 
          all_antecedents_here = false;
      }
      
      if (all_antecedents_here)
      {
        Resolution copy = r;        
        it = work_list.erase(it); // invalidates r;
        did_something = true;      
        do_resolution(copy);
      }
      else
        it++;
    }
  } while (did_something);
}

/*******************************************************************\
 
Function: QBV::do_resolution
 
  Inputs: resolution object, bool 
 
 Outputs: nothing
 
 Purpose: actually performs the resolution provided
 
\*******************************************************************/

void QBV::do_resolution( Resolution &resolution )
{
  // check resolution chain
  if (resolution.antecedents.empty()) // this is a root clause
  {
    if (!ignore_no_reason)
      throw Exception("Resolvent not justified (no antecedents).");
    qcnf.push_back(resolution.resolvent);
    if (resolution.newinx>=clause_map.size())
      clause_map.resize(resolution.newinx+1, 0);
    clause_map[resolution.newinx] = qcnf.size(); // this is 1 too high!
  }
  else if (resolution.antecedents.size()==1)
  {
    throw Exception("Single antecedent does not make sense.");
  }  
  else if (!resolution.star_resolvent)
  { 
    if (debug)
      std::cout << "DEBUG: " << resolution.newinx << ": supposed resolvent " << 
        resolution.resolvent.to_string() << std::endl;
        
    static std::list<unsigned> pivots;
    static DClause tclause, rclause;
    
    pivots.clear();       
    order(resolution.resolvent, resolution.antecedents, pivots);
    
    std::list<unsigned>::const_iterator it = resolution.antecedents.begin();
    tclause = qcnf[clause_map[*it] - 1];
    it++; 
    
    bool toggler = false;
    for (std::list<unsigned>::const_iterator pit = pivots.begin();
         it!=resolution.antecedents.end() && pit!=pivots.end();
         it++, pit++)
    {            
      if (!
        qcnf[clause_map[*it]-1].qresolve_into(
          (toggler)?rclause:tclause, 
          *pit,
          qcnf.quantification_levels,
          qcnf.quantifiers,
          (toggler)?tclause:rclause))
        throw Exception("Resolution failed.");
          
      toggler = !toggler;
    }    
    DClause &final = (toggler)?rclause:tclause;
    
    if (debug)
      std::cout << "DEBUG: " << resolution.newinx << ": " << 
        final.to_string() << std::endl;
    
    qcnf.push_back(final);
    if (resolution.newinx>=clause_map.size())
      clause_map.resize(resolution.newinx+1, 0);
    clause_map[resolution.newinx] = qcnf.size(); // this is 1 too high!
    
    if (debug)
      std::cout << "DEBUG: Inserted " << resolution.newinx << " as " << 
        qcnf.size()-1 << std::endl;
    
    // check that the given resolvent is equal
    if (resolution.resolvent.size()!=final.size())
      throw Exception("Calculated and supplied resolvent sizes do not match.");
    for (DClause::const_iterator it=final.begin();
         it!=final.end();
         it++)
    {
      if (!resolution.resolvent.contains(*it))
        throw Exception("Calculated and supplied resolvents differ.");
    }
  }
  else // *-resolvent
  {    
    qcnf.push_back(DClause());
    DClause &resolvent = qcnf.back();
    std::list<unsigned> doubles;
    for (std::list<unsigned>::const_iterator it = 
           resolution.antecedents.begin();
         it!=resolution.antecedents.end();
         it++)
    {
//      if ((clause_map[*it])==0)
//      {
//        std::stringstream s;
//        s << "BUG: Antecedent #" << *it << " not here for #" << 
//          resolution.newinx << std::endl;
//        throw Exception(s.str());
//      }      
      DClause &a = qcnf[clause_map[*it]-1];
      for (DClause::const_iterator cit = a.begin();
           cit!=a.end();
           cit++)
      {
        Literal l = *cit; l.invert();        
        if (resolvent.contains(l))
          doubles.push_back(l.variable());        
        resolvent.insert(*cit);
      }
    }
    for (std::list<unsigned>::const_iterator it = doubles.begin();
         it!=doubles.end();
         it++)
    {
      resolvent.remove(*it);
    }
    
    if (debug)
      std::cout << resolution.newinx << ": " << resolvent.to_string() << 
        std::endl;
        
    if (resolution.newinx>=clause_map.size())
      clause_map.resize(resolution.newinx+1, 0);
    clause_map[resolution.newinx] = qcnf.size(); // this is 1 too high!
  }
}

/*******************************************************************\
 
Function: QBV::extension_line
 
  Inputs: new variable number, certificate file object
 
 Outputs: nothing
 
 Purpose: processes an extension line
 
\*******************************************************************/

void QBV::extension_line( std::ifstream &certificate_file )
{
  unsigned newvar = read_unsigned(certificate_file);
  
  if (qcnf.is_defined(newvar))
    throw Exception("Extension variables must be fresh.");
  
  std::string &t = read_token(certificate_file);
  if (t=="I")
    extension_ite(newvar, certificate_file);    
  else if (t=="A")
    extension_and(newvar, certificate_file);
  else
    throw Exception("Unknown extension type supplied.");
}

/*******************************************************************\
 
Function: QBV::extension_ite
 
  Inputs: certificate file object
 
 Outputs: nothing
 
 Purpose: processes an ITE-extension line
 
\*******************************************************************/

void QBV::extension_ite( unsigned newvar, std::ifstream &certificate_file )
{
  signed x = read_signed(certificate_file);
  signed a = read_signed(certificate_file);
  signed b = read_signed(certificate_file);
  
  Literal xl( (x<0)?-x:x, (x<0) );
  Literal al( (a<0)?-a:a, (a<0) );
  Literal bl( (b<0)?-b:b, (b<0) );
  
  if (!qcnf.is_defined(xl.variable()))
    throw Exception("Extension uses undefined if-variable.");
  if (!qcnf.is_defined(al.variable()))
    throw Exception("Extension uses undefined then-variable.");
  if (!qcnf.is_defined(bl.variable()))
    throw Exception("Extension uses undefined else-variable.");
            
  // find quantification level of new variable
  unsigned maxlvl = qcnf.quantification_levels[xl.variable()];
  Quantifier maxq =  qcnf.quantifiers[xl.variable()];
  if (qcnf.quantification_levels[al.variable()] > maxlvl)
  {
    maxlvl = qcnf.quantification_levels[al.variable()];
    maxq = qcnf.quantifiers[al.variable()];
  }
  if (qcnf.quantification_levels[bl.variable()] > maxlvl)
  {
    maxlvl = qcnf.quantification_levels[bl.variable()];
    maxq = qcnf.quantifiers[bl.variable()];
  }
  if (! qcnf.new_variable_after(
          newvar, Quantifier::EXISTENTIAL, maxq.variable) )
    throw Exception(qcnf.last_error);
  
  CNF f;
  DClause c1, c2, c3, c4; // 4 clauses will be added to the qcnf.
  c1.insert(Literal(newvar, true));        
  c1.insert(al);
  c2.insert(Literal(newvar, true));
  c2.insert(xl);
  c2.insert(bl);
  c3.insert(Literal(newvar, false));        
  c4.insert(Literal(newvar, false));
  c4.insert(xl);
  xl.invert(); al.invert(); bl.invert();
  c1.insert(xl);
  c3.insert(xl);
  c3.insert(al);
  c4.insert(bl);
  qcnf.push_back(c1);       
  qcnf.push_back(c2);
  qcnf.push_back(c3);
  qcnf.push_back(c4);  
}

/*******************************************************************\
 
Function: QBV::extension_and
 
  Inputs: new variable number, certificate file object
 
 Outputs: nothing
 
 Purpose: processes an AND-extension line
 
\*******************************************************************/

void QBV::extension_and( unsigned newvar, std::ifstream &certificate_file )
{
  Literal newlit(newvar, false);
  DClause bigone;
  bigone.insert(newlit);
  
  signed x = read_signed(certificate_file);
  unsigned maxlvl = qcnf.quantification_levels[(x<0)?-x:x];
  Quantifier maxq = qcnf.quantifiers[(x<0)?-x:x];
  newlit.invert();
  while (x!=0) // n ...
  {
    Literal l((x<0)?-x:x, (x<0));
    
    DClause c;
    c.insert(newlit);
    c.insert(l);
    qcnf.push_back(c);
    
    l.invert();
    bigone.insert(l);
    
    if (qcnf.quantification_levels[l.variable()] > maxlvl)
    {
      maxlvl = qcnf.quantification_levels[l.variable()];
      maxq = qcnf.quantifiers[l.variable()];
    }
    
    x = read_signed(certificate_file);
  }        
  qcnf.push_back(bigone); // ... + 1 clauses will be added to the qcnf.
  
  if ( bigone.size()==1 ) // doesn't depend on anything
  {
    if (! qcnf.new_variable( newvar, Quantifier::EXISTENTIAL, true ) )
      throw Exception(qcnf.last_error);
  }   
  else 
  { 
    if (! qcnf.new_variable_after(
          newvar, Quantifier::EXISTENTIAL, maxq.variable) )
      throw Exception(qcnf.last_error);
  }
}

/*******************************************************************\
 
Function: QBV::conclusion_line
 
  Inputs: certificate file object
 
 Outputs: nothing
 
 Purpose: processes a conclusion line
 
\*******************************************************************/

bool QBV::conclusion_line( std::ifstream &certificate_file )
{
  if (!work_list.empty())
  {
    std::stringstream s;
    s << "The following delayed resolutions are missing antecedents " <<
      "at conclusion: " << std::endl;
    for (std::list<Resolution>::iterator it = work_list.begin();
         it!=work_list.end();
         it++)
    {
      s << it->newinx << ": ";
      for (std::list<unsigned>::const_iterator ait = it->antecedents.begin();
           ait!=it->antecedents.end();
           ait++)
      {
        s << *ait;
        if (clause_map[*ait]==0) s << "*";
        s << ", ";
      } 
      s << std::endl;     
    }
    throw Exception(s.str());
  }
    
  std::string &t = read_token(certificate_file);
  if (t=="INVALID" || t=="invalid")
  {
    unsigned ci = read_unsigned(certificate_file);
    DClause final_clause = qcnf.at(clause_map[ci]-1);
    final_clause.reduce(qcnf.quantification_levels, qcnf.quantifiers);
    return (final_clause.size()==0);
  }
  else if (t=="VALID" || t=="valid");
  {
    std::map<unsigned, signed> model;
    
    while (!certificate_file.eof())
    {
      unsigned exvar = read_unsigned(certificate_file);
      signed modelvar = read_signed(certificate_file);
      read_whitespace(certificate_file);
      
      if (qcnf.quantification_levels[exvar] <
          qcnf.quantification_levels[(modelvar<0)?-modelvar:modelvar])
      {
        std::stringstream s;
        s << "Model function (for " << exvar << " on lvl " << 
          qcnf.quantification_levels[exvar] << ") depends on variables " <<
          "with higher quantification levels (" << 
          ((modelvar<0)?-modelvar:modelvar) << " on " << 
          qcnf.quantification_levels[(modelvar<0)?-modelvar:modelvar] << ").";
        throw Exception(s.str());
      }  
      
      if (qcnf.quantifiers[exvar].type!=Quantifier::EXISTENTIAL)
        throw Exception("Universal variable in model.");
      
      model[exvar] = modelvar;
    }
        
    return testClausewise( model );
  }
}
      
/*******************************************************************\
 
Function: QBV::validate
 
  Inputs: string, string
 
 Outputs: true if the certificate was correct, false otherwise
 
 Purpose: validates a certificate file.
 
\*******************************************************************/

bool QBV::validate( const std::string &qbf, const std::string &qbc )
{
  if (test_trace) return validate_trace(qbc);
  
  std::string t;
  std::ifstream certificate_file(qbc.c_str());  
  
  if (!certificate_file.is_open())
    throw Exception("Cannot open file: '" + qbc +"'.");
  
  t = read_token(certificate_file);
  if (t!="QBCertificate")
    throw Exception("Missing header in file: '" + qbc + "'.");
    
//  t = read_token(certificate_file);
//  size_t i = filename.find_last_of("\\/");
//  if (i!=std::string::npos)
//    qbf_filename = filename.substr(0, i+1);
//  qbf_filename += t;
//  std::cout << "Validating against " << qbf_filename << std::endl;
  
  if (!qbf.empty())
  {
    std::ifstream qbf_file(qbf.c_str());
    if (!qbf_file)
      throw Exception("Couldn't open file " + qbf);
    if (!qcnf.read_qdimacs(qbf_file))
      throw Exception(qcnf.last_error);
  }
    
  clause_map.resize(qcnf.size()+1, 0);
  for (unsigned i=1; i<=qcnf.size(); i++)
  {
    clause_map[i]=i;
  }
  
  last_original_clause = qcnf.size() - 1;  
  
  if (debug)
  {
    unsigned qcount = 0;
    for (std::vector<QCNF::qscopet>::const_iterator it = 
           qcnf.quantification_sequence.begin();
         it!=qcnf.quantification_sequence.end();
         it++)
      qcount += it->size();
    std::cout << qcount << " quantifiers and " <<
      qcnf.size() << " clauses loaded." << std::endl;
  }
      
  while (certificate_file)
  {
    t = read_token(certificate_file);    
    if (t=="C" || t=="c") // Comment line
    {      
      while ( !certificate_file.eof() && 
              read_character(certificate_file)!='\n');
    } else if (t=="R" || t=="r") // Resolution line
    {      
      resolution_line(certificate_file);
    }
    else if (t=="E" || t=="e") // Extension Rules
    { 
      extension_line(certificate_file);
    }
    else if (t=="CONCLUDE" || t=="conclude")
    {
      return conclusion_line(certificate_file);
    }
    else
    {            
      // assume trace-style
      errno=0;  
      char *endp;
      unsigned val = strtoul(t.c_str(), &endp, 10);
      if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
      {    
        std::stringstream s;
        s << "Number conversion failed (" << t << ").";
        throw Exception(s.str());
      }
      trace_line(val, certificate_file);
    }
  }
  
  // At this point we should have read a conclusion line already
  throw Exception("Certificate incomplete.");
}

/*******************************************************************\
 
Function: QBV::testClausewise
 
  Inputs: model
 
 Outputs: true if the certificate was correct, false otherwise
 
 Purpose: checks if the provided model actually is a model.
 
\*******************************************************************/

bool QBV::testClausewise( std::map<unsigned, signed> &model )
{
  #ifdef __USE_MINISAT2
    SimpSolver solver;
  #else
    Solver solver;
  #endif
  
  while ((unsigned) solver.nVars()<=qcnf.max_variable_no)
    solver.newVar();
    
  // add function definitions
  for (unsigned i=last_original_clause+1; i<qcnf.size(); i++)
  {
    vec<Lit> c;    
    for (DClause::const_iterator lit = qcnf[i].begin();
         lit!=qcnf[i].end();
         lit++)
    {            
      c.push( Lit(lit->variable(), lit->sign()) );      
    }           
    solver.addClause(c);    
  }    
  
  // add model
  for (std::map<unsigned, signed>::const_iterator mit = model.begin();
       mit!=model.end();
       mit++)
  {    
    vec<Lit> c1, c2;
    c1.push( Lit(mit->first, false) );
    c1.push( Lit((mit->second<0)?-mit->second:mit->second, !(mit->second<0)) );
    
    c2.push( Lit(mit->first, true) ); 
    c2.push( Lit((mit->second<0)?-mit->second:mit->second, (mit->second<0)) );
    
    solver.addClause(c1);
    solver.addClause(c2);
  }  
  
  #ifdef __USE_MINISAT2
    solver.eliminate();
  #endif
  
  // test every clause.
  for (unsigned i=0; i<=last_original_clause; i++)
  {    
    vec<Lit> assumptions;
    for (DClause::const_iterator lit = qcnf[i].begin();
         lit!=qcnf[i].end();
         lit++)
    {
      assumptions.push( Lit( lit->variable(), ! lit->sign() ) );
    }
    
    if (!solver.okay()) throw Exception("Minisat not ready.");
    
    if (solver.solve(assumptions))
    {
      std::cout << "Counterexample found." << std::endl;
      std::cout << "Clause (#" << i << "): " << qcnf[i].to_string() << std::endl;
      if (debug)
      {
        std::cout << "Assignment to universals: ";
        for (std::vector<QCNF::qscopet>::const_iterator it = 
               qcnf.quantification_sequence.begin();
             it!=qcnf.quantification_sequence.end();
             it++)
        {
          if (it->front().type==Quantifier::UNIVERSAL)
          {
            for (QCNF::qscopet::const_iterator sit = it->begin();
                 sit!=it->end();
                 sit++)
            {
              std::cout << sit->variable << "=" << (solver.model[sit->variable]==l_True) << " ";
            }
          }
        }
        std::cout << std::endl;
      }
      return false;
    } // else all fine.          
  }
  
  return true; // no counterexample found
}

/*******************************************************************\
 
Function: QBV::read_unsigned
 
  Inputs: stream
 
 Outputs: unsigned
 
 Purpose: reads a token from a stream and tries to convert it to an 
          unsigned. Throws exceptions on errors.
 
\*******************************************************************/

unsigned QBV::read_unsigned( std::ifstream &in )
{
  std::string &t = read_token(in);
  //std::cout << "DEBUG: Converting " << t << std::endl;
  errno=0;  
  char *endp;
  unsigned val = strtoul(t.c_str(), &endp, 10);
  if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
  {    
    std::stringstream s;
    s << "Number conversion failed (" << t << ").";
    throw Exception(s.str());
  }
  return val;
}

/*******************************************************************\
 
Function: QBV::read_signed
 
  Inputs: stream
 
 Outputs: signed
 
 Purpose: reads a token from a stream and tries to convert it to a 
          signed integer. Throws exceptions on errors.
 
\*******************************************************************/

signed QBV::read_signed( std::ifstream &in )
{
  std::string &t = read_token(in);
  //std::cout << "DEBUG: Converting " << t << std::endl;
  errno=0;  
  char *endp;
  signed val = strtol(t.c_str(), &endp, 10);
  if (errno!=0 || !(*t.c_str()!='\0' && *endp=='\0'))
  {    
    std::stringstream s;
    s << "Number conversion failed (" << t << ").";
    throw Exception(s.str());
  }
  return val;
}

/*******************************************************************\
 
Function: QBV::read_token
 
  Inputs: stream
 
 Outputs: string
 
 Purpose: reads a token from a stream. Throws exceptions on errors.
 
\*******************************************************************/

std::string& QBV::read_token( std::ifstream &in )
{  
  char c;
  
  cur_token.clear();
  // pre-white
  c = in.peek();
  while ((c==' ' || c=='\n' || c=='\r') && !in.eof())
  {
    read_character(in);
    c = in.peek();
  }
  // token  
  while (c!=' ' && c!='\n' && c!='\r' && !in.eof())
  {
    cur_token+=read_character(in);
    c = in.peek();
  }
  
  //std::cout << "DEBUG: Read '" << cur_token << "'." << std::endl;
  
  return cur_token;
}

/*******************************************************************\
 
Function: QBV::read_character
 
  Inputs: stream
 
 Outputs: char
 
 Purpose: reads a character from a stream. Throws exceptions on errors.
 
\*******************************************************************/

char QBV::read_character( std::ifstream &in )
{
  char c;
  in.get(c);
  if (c=='\n' || c=='\r') 
  {
    cur_linenumber++;
    cur_colnumber=1;
  }
  else
    cur_colnumber++;
  return c;
}

/*******************************************************************\
 
Function: QBV::read_whitespace
 
  Inputs: stream
 
 Outputs: nothing
 
 Purpose: eats whitespace characters
 
\*******************************************************************/

void QBV::read_whitespace( std::ifstream &in )
{
  char c = in.peek();
  while (!in.eof() && (c==' ' || c=='\n' || c=='\r' || c=='\t'))
  { 
    read_character(in); 
    c=in.peek(); 
  }  
}

/*******************************************************************\
 
Function: QBV::validate_trace
 
  Inputs: string
 
 Outputs: true if the trace is correct, false otherwise.
 
 Purpose: validates a resolution trace.
 
\*******************************************************************/

bool QBV::validate_trace( const std::string &filename )
{
  std::ifstream trace_file(filename.c_str());
  
  clause_map.clear();
  
  while (!trace_file.eof())
  {
    read_whitespace(trace_file);
    if (trace_file.eof()) break;
    
    trace_line(read_unsigned(trace_file), trace_file);
  }
    
  if (qcnf.back().size()==0) // the last one must be empty then.
    return true;  
  
  trace_file.close();  
  return false;
}

/*******************************************************************\
 
Function: QBV::order
 
  Inputs: resolvent and antecedents
 
 Outputs: nothing
 
 Purpose: orders the antecedents so that trivial resolution can be done.
 
\*******************************************************************/

void QBV::order(  const DClause &resolvent, 
                  std::list<unsigned> &antecedents,
                  std::list<unsigned> &pivots ) const
{
  if (debug)
    std::cout << "DEBUG: ordering: " << antecedents.size() <<
      " antecedents." << std::endl;
      
  DClause assumptions = resolvent; // copy
  std::vector<DClause> clauses;
  std::vector<unsigned> clause_indexes;
  
  unsigned checksize = antecedents.size();
  
  //std::cout << "DEBUG: old order: ";
  for (std::list<unsigned>::const_iterator it = antecedents.begin();
       it!=antecedents.end();
       it++)
  {    
    if (*it>=clause_map.size() || clause_map[*it]==0)
      throw Exception("Inconsistent clause map.");
    clauses.push_back(qcnf.at(clause_map[*it]-1));
    clause_indexes.push_back(*it);
    
    if (debug)
      std::cout << "DEBUG: Added [" << *it << "] (" << clause_map[*it] << 
        ") = " << qcnf.at(clause_map[*it]-1).to_string() << std::endl;
  }
  
  antecedents.clear();
  
  unsigned last_dirty = 0;
  unsigned cur = 0;
  if (clauses.size()!=0)
  do
  {    
    for (DClause::const_iterator rit = assumptions.begin();
         rit!=assumptions.end();
         rit++)
    {
      Literal l = *rit; // assume -l is true 
      clauses[cur].remove(l);
      l.invert();
//      if (clauses[cur].contains(l))
//      {
//        //std::cout << "Satisfied clause" << std::endl;        
//        goto next;
//      }
    }
    
    unsigned virtual_clause_size=0;
    Literal virtual_unit(0,0);
        
    for (DClause::const_iterator cit = clauses[cur].begin();
         cit!=clauses[cur].end();
         cit++)
    {
      if (qcnf.quantifiers[cit->variable()].type==Quantifier::EXISTENTIAL)
      {
        virtual_clause_size++;
        virtual_unit = *cit;
      }
    }
    
    if (virtual_clause_size==0)
    {      
      antecedents.push_back(clause_indexes[cur]);      
    }
    else
    {
      virtual_unit.invert();
      if (virtual_clause_size==1 &&
          assumptions.insert(virtual_unit))    // new unit
      {
        // std::cout << "Found assumption: " << u.to_dimacs() << std::endl;
        antecedents.push_back(clause_indexes[cur]);        
        pivots.push_back(virtual_unit.variable());
        last_dirty = cur;
      }      
    }
    
//    next:
    cur++;
    if (cur==clauses.size()) cur = 0; //wrap    
  } while (cur!=last_dirty);
  
  if (antecedents.size()==checksize-1)
  {
    if (cur==0) //one back 
      cur = clauses.size()-1;
    else
      cur--;
    antecedents.push_back(clause_indexes[cur]);
  }
  else if (antecedents.size()!=checksize)
  {
    for (std::list<unsigned>::const_iterator it = antecedents.begin();
         it!=antecedents.end();
         it++)
    {
      if (*it>=clause_map.size() || clause_map[*it]==0)
        throw Exception("Inconsistent clause map.");
      std::cout << *it << ": " << qcnf.at(clause_map[*it]-1).to_string() << 
        std::endl;
    }
    throw Exception("BUG: Lost antecedents while reordering.");
  }
    
  antecedents.reverse();
  pivots.reverse();
  
  if (debug)
  {
    std::cout << "DEBUG: new order: ";
    for (std::list<unsigned>::const_iterator it = antecedents.begin();
         it!=antecedents.end();
         it++)
    {
      std::cout << *it << ", ";
    }    
    std::cout << std::endl;
    std::cout << "DEBUG: pivots: ";
    for (std::list<unsigned>::const_iterator it = pivots.begin();
         it!=pivots.end();
         it++)
    {
      std::cout << *it << ", ";
    }
    std::cout << std::endl;
  }
}
